#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#define IMIN(a, b) ((a) < (b) ? (a) : (b))

double wtime()
{
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return ts.tv_sec + ts.tv_nsec * 1E-9;
}

void dgemv(double *a, double *b, double *c, int m, int n)
{
    for (int i = 0; i < m; i++) {
        for (int j = 0; j < n; j++)
            c[i] += a[i * n + j] * b[j];
    }
}

void dgemv_block(double *a, double *b, double *c, int m, int n)
{
    int bs = 8;
    for (int i = 0; i < m; i += bs) {
        for (int j = 0; j < n; j += bs) {
            for (int ii = i; ii < IMIN(i + bs, m); ii++) {
                for (int jj = j; jj < IMIN(j + bs, n); jj++) {
                    c[ii] += a[ii * n + jj] * b[jj];
                }
            }
        }
    }
}

double run(int m, int n)
{
    double *a, *b, *c;
    a = aligned_alloc(64, sizeof(*a) * m * n);
    b = aligned_alloc(64, sizeof(*b) * n);
    c = aligned_alloc(64, sizeof(*c) * m);
    for (int i = 0; i < m; i++) {
        for (int j = 0; j < n; j++)
            a[i * n + j] = i + j;
        c[i] = 0;
    }
    for (int j = 0; j < n; j++)
        b[j] = j;

    double t = wtime();
    dgemv(a, b, c, m, n);
    //dgemv_block(a, b, c, m, n);
    t = wtime() - t;

    free(a);
    free(b);
    free(c);
    return t;
}

int main()
{
    int nreps = 100;
    printf("# [m]  [time, sec]   [MFLOPS]\n");
    int n = 1000;
    for (int m = 1000; m <= 10000; m += 1000) {
        double t = 0;
        for (int i = 0; i < nreps; i++)
             t += run(m, n);
        t /= nreps;
        printf(" %-6d  %-12.6f  %-12.2f\n", m, t, 2.0 * m * n * 1e-6 / t);
    }

    return 0;
}

